/*=============================================================================
Turing Designer
de_pad.cpp - Designer Drawpad

  Copyright © Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2005
  -- License restrictions apply (GPL v2)

  This file is part of Turing Designer.
  Turing Desigenr is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; however ONLY version 2 of the License.

  Turing Designer is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program kit; if not, write to:
  Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.

  -- For details see the COPYING file
=============================================================================*/
#include <math.h>
#include <stdio.h>
#include <wx/wxprec.h>
#ifndef WX_PRECOMP
#    include <wx/wx.h>
#endif
#include <wx/statline.h>
#include "de_data.hpp"
#include "de_pad.hpp"
#include "gr_data.hpp"
#include "xf_wxsu.hpp"
#include "xf_xml.hpp"

#define ARROW_LENGTH 15 // length of arrow sidelines, in pixels
#define ARROW_FLAP   20 // angle between the baseline and one sideline (degrees)

enum {
    SE_SELECT   = 1 << 0,
    SE_DESELECT = 1 << 1,

    // Drawing area
    DA_WIDTH  = 4096,
    DA_HEIGHT = DA_WIDTH,
};

union bbox {
    int i[4];
    struct {
        int left, top, right, bottom;
    } n;
};

class Dlg_LinkDest : public wxDialog {
  public: // functions
    Dlg_LinkDest(wxWindow *);
    void results(char *, char *, char *) const;

  private: // variables
    wxTextCtrl *tx_ichar, *tx_ochar;
    wxChoice *ch_mov;
};

class Dlg_DelLinkDest : public wxDialog {
  public: // functions
    Dlg_DelLinkDest(wxWindow *);
    char result(void) const;

  private: // variables
    wxTextCtrl *tx_ichar;
};

class Dlg_INodeDest : public wxDialog {
  public: // functions
    Dlg_INodeDest(wxWindow *);
    wxString result(void) const;

  private: // variables
    wxTextCtrl *tx_file;
};

static wxPen PEN_Orange(wxColour(255, 128, 0));

BEGIN_EVENT_TABLE(Drawpad, wxPanel)
    EVT_MENU(wxID_DELETE, Drawpad::Delete)
    EVT_ENTER_WINDOW    (Drawpad::Mouse_Enter)
    EVT_LEFT_DOWN       (Drawpad::Mouse_LeftDown)
    EVT_LEFT_UP         (Drawpad::Mouse_LeftUp)
    EVT_MOTION          (Drawpad::Mouse_Motion)
    EVT_ERASE_BACKGROUND(Drawpad::Window_Erase)
    EVT_PAINT           (Drawpad::Window_Paint)
    EVT_SIZE            (Drawpad::Window_Resize)
    EVT_SCROLLWIN       (Drawpad::Window_Scroll)
END_EVENT_TABLE()

//-----------------------------------------------------------------------------
Drawpad::Drawpad(wxWindow *parent, TGraph *tg) :
    wxPanel(parent, wxID_ANY, wxDefaultPosition, wxDefaultSize,
     wxSUNKEN_BORDER | wxFULL_REPAINT_ON_RESIZE)
{
    mouse_motion_oc = false;
    mouse_prev_sel  = false;
    mouse_inode     = -1;
    mode            = DWEXT_MARK;
    graph           = tg;
    vpx = vpy = vpw = vph = 0; // viewport
    SetBackgroundColour(*wxWHITE);
    return;
}

void Drawpad::Delete(wxCommandEvent &ev) {
    // Delete all selected nodes
    int w = graph->last_idx();
    XVNode *node;

    while((node = dynamic_cast<XVNode *>(graph->walk_rev(&w))) != NULL) {
        if(node->sel) {
            delete_targets(node);
            graph->del(w + 1);
        }
    }
    return;
}

void Drawpad::ext_action(unsigned int a) {
    // Called from class Simulator
    mode = a;
    switch_crosshair();
    return;
}

//-----------------------------------------------------------------------------
void Drawpad::Mouse_Enter(wxMouseEvent &ev) {
    switch_crosshair();
    return;
}

void Drawpad::Mouse_Motion(wxMouseEvent &ev) {
    if(!ev.m_leftDown || mode != DWEXT_MARK || mouse_inode < 0) { return; }

    XVNode *pnode = dynamic_cast<XVNode *>((*graph)[mouse_inode]);
    mouse_motion_oc = true;
    pnode->x += ev.GetX() - mouse_prev_pos.x; // apply delta movement
    pnode->y += ev.GetY() - mouse_prev_pos.y;
    mouse_prev_pos = ev.GetPosition();
    select_node(mouse_inode);
    Refresh();
    return;
}

void Drawpad::Mouse_LeftUp(wxMouseEvent &ev) {
    if(mode != DWEXT_MARK || mouse_inode < 0) { return; }

    /* Kept for historical purposes:
    >>
    >> if(mouse_motion_oc) {
    >>     if(!mouse_prev_sel) {
    >>         select_node(mouse_inode, SE_DESELECT);
    >>     } else {
    >>         select_node(mouse_inode, SE_SELECT);
    >>     }
    >> } else {
    >>     if(!mouse_prev_sel) {
    >>         // already selected
    >>     } else {
    >>         select_node(mouse_inode, SE_DESELECT);
    >>     }
    >> }
    >>
    Nicely shortened/redone using XOR: */

    if(mouse_motion_oc ^ mouse_prev_sel) {
        select_node(mouse_inode, SE_DESELECT);
    } else if(mouse_prev_sel) {
        select_node(mouse_inode, SE_SELECT);
    }

    mouse_motion_oc = false;
    mouse_inode = -1;
    return;
}

void Drawpad::Mouse_LeftDown(wxMouseEvent &ev) {
    /* If a sequence is complete, the object is drawn _now_, in
    Mouse_LeftDown(), because indow_Paint() is not directly called
    afterwards. As for DWEXT_..., at least one object is always visible (either
    A+B, or just B), thus, we do not need any viewport optimization here. */

    int px = ev.GetX(), py = ev.GetY(), tx = px + vpx, ty = py + vpy,
     node_num = -1;
    wxPaintDC dc(this);
    XVNode *node;
    int node_int;

    dc.BeginDrawing();

    switch(mode) {
        case DWEXT_MARK:
            if((node_int = find_node_i(tx, ty)) >= 0) {
                node = dynamic_cast<XVNode *>((*graph)[node_int]);
                mouse_motion_oc = false;
                mouse_prev_pos  = ev.GetPosition();
                mouse_prev_sel  = node->sel;
                mouse_inode     = node_int;
                if(!node->sel) {
                    select_node(mouse_inode, SE_SELECT);
                }
            }
            break;
        case DWEXT_SNODE: {
            XSNode *snode = new XSNode(tx, ty);
            draw_node(snode, graph->add(snode), snode->x - vpx,
             snode->y - vpy, &dc);
            break;
        }
        case DWEXT_INODE: {
            Dlg_INodeDest dlg(this);
            if(dlg.ShowModal() != wxID_OK) { return; }

            XINode *inode = new XINode(tx, ty);
            wxString res  = dlg.result();
            inode->ext    = HX_strdup(res);
            inode->graph  = xg_load(inode->ext);
            draw_node(inode, graph->add(inode), inode->x - vpx,
             inode->y - vpy, &dc);
            break;
        }
        case DWEXT_LNODE:
            node = new XLNode(tx, ty);
            draw_node(node, graph->add(node), node->x - vpx,
             node->y - vpy, &dc);
            break;
        case DWEXT_LINK:
            // A link requires two nodes, here is the first.
            if((node = find_node(tx, ty)) != NULL) {
                templink = node;
                mode = DWEXT_LINK1;
            }
            break;
        case DWEXT_LINK1:
            // Second...
            if((node_int = find_node_i(tx, ty)) >= 0) {
                do_link(templink, node_int, &dc);
                mode = DWEXT_LINK;
            }
            break;
        case DWEXT_LINKD:
            if((node = find_node(tx, ty)) != NULL) {
                do_del_link(node, &dc);
            }
            break;
        case DWEXT_ENTRYPT:
            if((node_int = find_node_i(tx, ty)) >= 0) {
                graph->entrypoint = node_int;
                mode = DWEXT_MARK;
            }
            break;
        default:
            fprintf(stderr, "%s: Invalid mode!\n", __PRETTY_FUNCTION__);
            break;
    }

    dc.EndDrawing();
    Refresh();
    switch_crosshair();
    return;
}

void Drawpad::Window_Erase(wxEraseEvent &ev) {
    /* Window is automatically erased, no need to do anything here; but the
    function needs to be present, otherwise _nothing_ happens. */
    return;
}

void Drawpad::Window_Paint(wxPaintEvent &ev) {
    // Redraw certain parts of the Drawpad panel

    wxRegionIterator iter(GetUpdateRegion());
    wxPaintDC dc(this);
    XVNode *node;

    dc.BeginDrawing();

    while(iter) {
        int num = 0, tx = iter.GetX() + vpx, ty = iter.GetY() + vpy,
         iw = iter.GetWidth(), ih = iter.GetHeight();

        while((node = dynamic_cast<XVNode *>(graph->walk(&num))) != NULL) {
            if(node->x + node->radius >= tx &&
             node->x - node->radius <= tx + iw &&
             node->y + node->radius >= ty &&
             node->y - node->radius <= ty + ih) {
                draw_node(node, num - 1, node->x - vpx, node->y - vpy, &dc);
            }
            switch(node->type()) {
                case TYPE_SNODE:
                    draw_arrows(dynamic_cast<XSNode *>(node),
                     tx, ty, iw, ih, &dc);
                    break;
                case TYPE_INODE:
                    draw_arrows(dynamic_cast<XINode *>(node),
                     tx, ty, iw, ih, &dc);
                    break;
                case TYPE_LNODE:
                    draw_arrows(dynamic_cast<XLNode *>(node),
                     tx, ty, iw, ih, &dc);
                    break;
                default:
                    fprintf(stderr, "%s: Invalid node type!\n",
                     __PRETTY_FUNCTION__);
                    break;
            }
        }
        ++iter;
    }

    dc.EndDrawing();
    return;
}

void Drawpad::Window_Resize(wxSizeEvent &ev) {
    // The window has been resized. Adjust viewport variables and scrollbars.
    vpw = ev.GetSize().GetX();
    vph = ev.GetSize().GetY();
    SetScrollbar(wxHORIZONTAL, vpx, vpw, DA_WIDTH);
    SetScrollbar(wxVERTICAL,   vpy, vph, DA_HEIGHT);
    thx = vpx;
    thy = vpy;
    ev.Skip();
    return;
}

void Drawpad::Window_Scroll(wxScrollWinEvent &ev) {
    /* Scrolling is just straightforward reproduction code (= it is always the
    same code base). Taken from wxWidgets/demos/life/life.cpp
    (LifeCanvas::OnScroll()). */

    WXTYPE type   = ev.GetEventType();
    int orient    = ev.GetOrientation();
    int pos       = ev.GetPosition();
    int scrollinc = 0;

    if(type == wxEVT_SCROLLWIN_TOP) {
        if(orient == wxHORIZONTAL) {
            scrollinc = -vpw;
        } else {
            scrollinc = -vph;
        }
    } else if(type == wxEVT_SCROLLWIN_BOTTOM) {
        if(orient == wxHORIZONTAL) {
            scrollinc = vpw;
        } else {
            scrollinc = vph;
        }
    } else if(type == wxEVT_SCROLLWIN_LINEUP) {
        scrollinc = -1;
    } else if(type == wxEVT_SCROLLWIN_LINEDOWN) {
        scrollinc = 1;
    } else if(type == wxEVT_SCROLLWIN_PAGEUP) {
        scrollinc = -10;
    } else if(type == wxEVT_SCROLLWIN_PAGEDOWN) {
        scrollinc = 10;
    } else if(type == wxEVT_SCROLLWIN_THUMBTRACK) {
        if(orient == wxHORIZONTAL) {
            scrollinc = pos - thx;
            thx = pos;
        } else {
            scrollinc = pos - thy;
            thy = pos;
        }
    } else if(type == wxEVT_SCROLLWIN_THUMBRELEASE) {
        thx = vpx;
        thy = vpy;
    }

#if !defined(__WXGTK__) && !defined(__WXMOTIF__)
    /* wxGTK and wxMotif update the thumb automatically (wxMSW does not);
    so reset it back as we always want it to be in the same position. */
    if(type != wxEVT_SCROLLWIN_THUMBTRACK) {
        SetScrollbar(wxHORIZONTAL, vpx, vpw, DA_WIDTH);
        SetScrollbar(wxVERTICAL,   vpy, vph, DA_HEIGHT);
    }
#endif

    if(scrollinc == 0) {
        return;
    }

    // Scroll the window and adjust the viewport
    if(orient == wxHORIZONTAL) {
        vpx += scrollinc;
        ScrollWindow(-scrollinc, 0, NULL);
    } else {
        vpy += scrollinc;
        ScrollWindow(0, -scrollinc, NULL);
    }
    return;
}

void Drawpad::delete_targets(VNode *victim) {
    /* Deletes all targets pointing to VICTIM.
    Called from: Edit_Delete()
    */

    VNode *node;
    int w = 0;

    while((node = graph->walk(&w)) != NULL) {
        switch(node->type()) {
            case TYPE_SNODE: {
                SNode *snode = dynamic_cast<SNode *>(node);
                void *travp = HXbtrav_init(snode->jmp, NULL);
                struct HXbtree_node *bt;

                while((bt = HXbtraverse(travp)) != NULL) {
                    ATarget *target = reinterpret_cast<ATarget *>(bt->data);
                    if(target->next >= 0 && (*graph)[target->next] == victim) {
                        snode->del(bt->key);
                    }
                }
                break;
            }
            case TYPE_INODE: {
                INode *inode = dynamic_cast<INode *>(node);
                if(inode->next >= 0 && (*graph)[inode->next] == victim) {
                    inode->next = -1;
                }
                break;
            }
            case TYPE_LNODE: {
                LNode *lnode = dynamic_cast<LNode *>(node);
                if(lnode->next >= 0 && (*graph)[lnode->next] == victim) {
                    lnode->next = -1;
                }
                break;
            }
            default:
                fprintf(stderr, "Drawpad::delete_targets(): Illegal node type\n");
                break;
        }
    }
    Refresh();
    return;
}

void Drawpad::do_del_link(XVNode *node, wxDC *dc) {
    // Called from: Mouse_LeftDown()
    switch(node->type()) {
        case TYPE_SNODE: {
            Dlg_DelLinkDest dlg(this);
            if(dlg.ShowModal() != wxID_OK) { return; }
            dynamic_cast<XSNode *>(node)->del(dlg.result());
            break;
        }
        case TYPE_INODE: {
            XINode *inode = dynamic_cast<XINode *>(node);
            inode->next = -1;
            break;
        }
        case TYPE_LNODE: {
            XLNode *lnode = dynamic_cast<XLNode *>(node);
            lnode->next = -1;
            break;
        }
        default:
            fprintf(stderr, "%s: Illegal node type!\n", __PRETTY_FUNCTION__);
            break;
    }
    Refresh();
    return;
}

void Drawpad::do_link(XVNode *from, int to, wxDC *dc) {
    // Called from: Mouse_LeftDown()
    switch(from->type()) {
        case TYPE_SNODE: {
            Dlg_LinkDest dlg(this);
            char in, out, mov;

            if(dlg.ShowModal() != wxID_OK) { break; }
            dlg.results(&in, &out, &mov);
            dynamic_cast<XSNode *>(from)->put(in, out, mov, to);
            draw_arrow(from, dynamic_cast<XVNode *>((*graph)[to]), dc);
            break;
        }
        case TYPE_INODE:
            dynamic_cast<XINode *>(from)->next = to;
            draw_arrow(from, dynamic_cast<XVNode *>((*graph)[to]), dc);
            break;
        case TYPE_LNODE:
            dynamic_cast<XLNode *>(from)->next = to;
            draw_arrow(from, dynamic_cast<XVNode *>((*graph)[to]), dc);
            break;
        default:
            fprintf(stderr, "%s: Illegal node type!\n", __PRETTY_FUNCTION__);
            break;
    }
    return;
}

void Drawpad::draw_arrow(XVNode *ap, XVNode *bp, wxDC *dc, int *mx,
 int *my) const
{
#define ARROW_FLAP_RAD ((ARROW_FLAP) * M_PI / 180)
    /* Draw an arrow between two nodes. It is is not as difficult as Modrow
    said... just requires a lot of initial thinking on paper.
    Called from: Mouse_LeftDown(), draw_arrows<*>()
    */
    double alpha, beta;
    int pax, pay, pbx, pby;
    int dx = bp->x - ap->x, dy = bp->y - ap->y;

    // Return center coordinates for SNode arrow description
    if(mx != NULL) { *mx = ap->x + dx / 2; }
    if(my != NULL) { *my = ap->y + dy / 2; }

    alpha = asin(dx / sqrt(pow(dx, 2) + pow(dy, 2)));
    if(dy >= 0) { alpha = M_PI - alpha; }

    // Baseline
    double sin_alpha = sin(alpha), cos_alpha = cos(alpha);
    pax = (int)(ap->x + sin_alpha * ap->radius) - vpx,
    pay = (int)(ap->y - cos_alpha * ap->radius) - vpy,
    pbx = (int)(bp->x - sin_alpha * bp->radius) - vpx,
    pby = (int)(bp->y + cos_alpha * bp->radius) - vpy;

    dc->SetPen(wxPen(*wxBLACK, 2));
    dc->DrawLine(pax, pay, pbx, pby);

    // Sidelines
    beta = M_PI - alpha - ARROW_FLAP_RAD; // left (clockwise) side
    dc->DrawLine(pbx, pby, (int)(pbx - sin(beta) * ARROW_LENGTH),
     (int)(pby - cos(beta) * ARROW_LENGTH));

    beta = alpha - ARROW_FLAP_RAD; // right (counter-clockwise) side
    dc->DrawLine(pbx, pby, (int)(pbx - sin(beta) * ARROW_LENGTH),
     (int)(pby + cos(beta) * ARROW_LENGTH));
    return;
#undef ARROW_FLAP_RAD
}

/* draw_arrows<*>() is an optimization prefunction: Only draw the arrow if the
update region is in its path. Since path pixel calculation would be a lot more
difficult, we only check whether the update region is within the _bounding box_
of the arrow path/line.

Called from: Window_Paint()
*/

void Drawpad::draw_arrows(XSNode *node, int ax, int ay, int aw, int ah,
 wxDC *dc) const
{
    /* An SNode can have multiple targets -- step through each and draw an
    arrow, as well as the target description. */

    void *travp = HXbtrav_init(node->jmp, NULL);
    struct HXbtree_node *nd;
    union bbox b;

    while((nd = HXbtraverse(travp)) != NULL) {
        ATarget *target = reinterpret_cast<ATarget *>(nd->data);
        XVNode *other   = dynamic_cast<XVNode *>((*graph)[target->next]);

        if(other == NULL) {
            fprintf(stderr,
               "%s: Incomplete jumptable!\n"
               "There is nothing wrong with that, but it should not happen.\n",
               __PRETTY_FUNCTION__
            );
            continue;
        }

        find_bounds(node, other, &b);
        if(!(ax <= b.n.right && ay <= b.n.bottom) &&
         !(ax + aw >= b.n.left && ay + ah >= b.n.top)) {
            continue;
        }

        /* May look weird if the arrow shows en direction NOT(EAST, WEST)
        because the text is not rotated to go along the arrow line. */
        wxCoord tx_width, tx_height;
        wxString str, stmp;
        int mid_x, mid_y;
        char key = reinterpret_cast<int>(nd->key);

        if(key != 0) { stmp.Printf("'%c',", key); }
        else         { stmp.Printf("-,"); }
        str += stmp;

        if(target->out != 0) { stmp.Printf("'%c',", target->out); }
        else                 { stmp.Printf("-,"); }
        str += stmp;

        stmp.Printf("%c", (target->mov != 0) ? target->mov : '-');
        str += stmp;

        draw_arrow(node, other, dc, &mid_x, &mid_y);
        dc->GetTextExtent(str, &tx_width, &tx_height);
        dc->SetTextForeground(*wxBLUE);
        dc->DrawText(str, mid_x - tx_width / 2 - vpx,
         mid_y - tx_height / 2 - vpy);
    }
    return;
}

template<class Object> void Drawpad::draw_arrows(Object *node, int x, int y,
 int w, int h, wxDC *dc) const
{
    /* Generic arrow drawing function for single-target nodes. This had to be a
    templated function because the base class XVNode does not have a "next"
    member. (And won't get one either.) */

    if(node->next < 0) { return; }

    XVNode *other = dynamic_cast<XVNode *>((*graph)[node->next]);
    union bbox b;

    find_bounds(node, other, &b);
    if((x <= b.n.right && y <= b.n.bottom) ||
     (x + w >= b.n.left && y + h >= b.n.top)) {
        draw_arrow(node, other, dc);
    }
    return;
}

void Drawpad::draw_node(XVNode *node, int node_int, int tx, int ty,
 wxDC *dc) const
{
    /* Draw a shape according to its parameters (set during
    construction of the node).
    Called from: Mouse_LeftDown(), Window_Paint(), select_node()
    */

    dc->SetPen(*node->pen);
    dc->SetBrush(*node->brush);

    switch(node->shape) {
        case DWSHP_CIRCLE:
            dc->DrawCircle(tx, ty, node->radius);
            break;
        case DWSHP_RECT:
            dc->DrawRectangle(tx - node->radius, ty - node->radius,
             node->radius * 2, node->radius * 2);
            break;
        default:
            fprintf(stderr, "%s: Illegal node shape\n", __PRETTY_FUNCTION__);
            break;
    }

    if(node->type() == TYPE_SNODE) {
        /* Additionally, if it is an SNode, output the state number, too. Note
        that even though every node has an array index within TGraph->node,
        only SNodes have a meaningful state. INodes and LNodes do not have a
        state, i.e. they are just a tunnel in the TuringEngine execution
        unit. */

        wxCoord tx_width, tx_height;
        wxString str;

        str.Printf("%u", node_int);
        dc->GetTextExtent(str, &tx_width, &tx_height);
        dc->SetTextForeground(*wxBLACK);
        dc->DrawText(str, node->x - tx_width / 2 - vpx,
         node->y - tx_height / 2 - vpy);
    } else if(node->type() == TYPE_INODE) {
        XINode *inode = dynamic_cast<XINode *>(node);
        if(inode->ext != NULL) {
            wxCoord tx_width, tx_height;
            wxString str;
            str += inode->ext;
            if(inode->graph == NULL) {
                str += " (error)";
            }
            dc->GetTextExtent(str, &tx_width, &tx_height);
            dc->SetTextForeground(*wxBLACK);
            dc->DrawText(str, node->x - tx_width / 2 - vpx,
             node->y - tx_height / 2 - vpy);
        }
    }

    return;
}

union bbox *Drawpad::find_bounds(XVNode *node, XVNode *other,
 union bbox *b) const
{
    /* Calculate the bounding box of the two nodes
    Returns:     the bbox we were given
    Called from: draw_arrows<*>()
    */

    if(node->x < other->x) {
        b->n.left   = node->x;
        b->n.right  = other->x;
    } else {
        b->n.left   = other->x;
        b->n.right  = node->x;
    }
    if(node->y < other->y) {
        b->n.top    = node->y;
        b->n.bottom = other->y;
    } else {
        b->n.top    = other->y;
        b->n.bottom = node->y;
    }
    return b;
}

XVNode *Drawpad::find_node(int x, int y) const {
    /* Find the node at position (x,y).
    Returns:     Pointer to node if found, otherwise NULL
    Called from: Mouse_LeftDown()
    */

    int w = graph->last_idx();
    XVNode *node;

    while((node = dynamic_cast<XVNode *>(graph->walk_rev(&w))) != NULL) {
        switch(node->shape) {
            case DWSHP_CIRCLE:
                // Hey, Pythargoras stuffage!
                if(sqrt(pow(node->x - x, 2) + pow(node->y - y, 2)) <
                 node->radius) {
                    return node;
                }
                break;
            case DWSHP_RECT:
                /* Add 1 to each border, since a rectangle with "radius" 5 is
                actually 11 pixels wide/tall. (See draw_node().) */
                if(node->x >= x - node->radius + 1 &&
                 node->x <= x + node->radius - 1 &&
                 node->y >= y - node->radius + 1 &&
                 node->y <= y + node->radius - 1) {
                    return node;
                }
                break;
            default:
                fprintf(stderr, "%s: Illegal node shape\n",
                 __PRETTY_FUNCTION__);
                break;
        }
    }
    return NULL;
}

int Drawpad::find_node_i(int x, int y) const {
    /* Find the node at position (x,y).
    Returns:     Pointer to node if found, otherwise NULL
    Called from: Mouse_LeftDown()
    */

    int w = graph->last_idx();
    XVNode *node;

    while((node = dynamic_cast<XVNode *>(graph->walk_rev(&w))) != NULL) {
        switch(node->shape) {
            case DWSHP_CIRCLE:
                // Hey, Pythargoras stuffage!
                if(sqrt(pow(node->x - x, 2) + pow(node->y - y, 2)) <
                 node->radius) {
                    return w + 1;
                }
                break;
            case DWSHP_RECT:
                /* Add 1 to each border, since a rectangle with "radius" 5 is
                actually 11 pixels wide/tall. (See draw_node().) */
                if(node->x >= x - node->radius + 1 &&
                 node->x <= x + node->radius - 1 &&
                 node->y >= y - node->radius + 1 &&
                 node->y <= y + node->radius - 1) {
                    return w + 1;
                }
                break;
            default:
                fprintf(stderr, "%s: Illegal node shape\n",
                 __PRETTY_FUNCTION__);
                break;
        }
    }
    return -1;
}

void Drawpad::select_node(unsigned int node_int, unsigned int flags) {
    /* Changes the state and graphics of a node according to flags.
    Called from: Mouse_Motion(), Mouse_LeftUp(), Mouse_LeftDown()
    */
    XVNode *node = dynamic_cast<XVNode *>((*graph)[node_int]);

    if(flags & SE_SELECT) {
        if(node->o_brush == NULL) { node->o_brush = node->brush; }
        if(node->o_pen == NULL)   { node->o_pen = node->pen; }
        node->sel   = true;
        node->brush = wxBLACK_BRUSH;
        node->pen   = &PEN_Orange;
    } else if(flags & SE_DESELECT) {
        node->sel     = false;
        node->brush   = node->o_brush;
        node->pen     = node->o_pen;
        node->o_brush = NULL;
        node->o_pen   = NULL;
    }

    wxPaintDC dc(this);
    dc.BeginDrawing();
    draw_node(node, node_int, node->x - vpx, node->y - vpy, &dc);
    dc.EndDrawing();
    return;
}

void Drawpad::switch_crosshair(void) {
    /* Switch the crosshair when entering/leaving the Drawpad. When leaving the
    drawpad, wxSTANDARD_CURSOR is always used, so no need to call
    switch_crosshair() upon leave.
    Called from: Mouse_Enter(), Mouse_LeftDown()
    */

    switch(mode) {
        case DWEXT_MARK:
            SetCursor(*wxSTANDARD_CURSOR);
            break;
        case DWEXT_SNODE:
        case DWEXT_INODE:
        case DWEXT_LNODE:
        case DWEXT_LINK:
        case DWEXT_LINK1:
        case DWEXT_LINKD:
        case DWEXT_ENTRYPT:
            SetCursor(*wxCROSS_CURSOR);
            break;
    }
    return;
}

//=============================================================================
Dlg_LinkDest::Dlg_LinkDest(wxWindow *parent) :
    wxDialog(parent, wxID_ANY, "SNode Link Target", wxDefaultPosition,
     wxDefaultSize, COMMON_FRAME_FLAGS)
{
    static const wxString mov_ch[] = {"None", "Left", "Right", "Error"};
    wxBoxSizer *vp = new wxBoxSizer(wxVERTICAL),
     *hp = new wxBoxSizer(wxHORIZONTAL);
    wxFlexGridSizer *sp = new wxFlexGridSizer(2);

    vp->Add(sp, 1, wxALIGN_CENTER | wxGROW | wxALL, 3);
    vp->Add(new wxStaticLine(this, wxID_ANY), 0, wxGROW | wxTOP | wxBOTTOM, 5);
    vp->Add(hp, 0, wxALIGN_RIGHT);

    tx_ichar = new wxTextCtrl(this, wxID_ANY, wxEmptyString);
    tx_ochar = new wxTextCtrl(this, wxID_ANY, wxEmptyString);
    ch_mov   = new wxChoice(this, wxID_ANY, wxDefaultPosition, wxDefaultSize,
               sizeof(mov_ch) / sizeof(typeof(*mov_ch)),
               (const wxString *)mov_ch);
    tx_ichar->SetMaxLength(1);
    tx_ochar->SetMaxLength(1);
    ch_mov->SetSelection(0);

    sp->Add(new wxStaticText(this, wxID_ANY, "Input char"), 0, wxACV | wxALL, 3);
    sp->Add(tx_ichar, 1, wxGROW | wxACV | wxALL, 3);
    sp->Add(new wxStaticText(this, wxID_ANY, "Output char"), 0, wxACV | wxALL, 3);
    sp->Add(tx_ochar, 1, wxGROW | wxACV | wxALL, 3);
    sp->Add(new wxStaticText(this, wxID_ANY, "Movement after write"), 0, wxACV | wxALL, 3);
    sp->Add(ch_mov, 1, wxGROW | wxACV | wxALL, 3);
    sp->AddGrowableCol(1);

    hp->Add(new wxButton(this, wxID_OK, "&OK"), 0, wxALL, 5);
    hp->Add(new wxButton(this, wxID_CANCEL, "&Cancel"), 0, wxALL, 5);

    SetSizer(vp);
    vp->SetSizeHints(this);
    return;
}

void Dlg_LinkDest::results(char *in, char *out, char *mov) const {
    if(in  != NULL) { *in  = tx_ichar->GetValue().GetChar(0); }
    if(out != NULL) { *out = tx_ochar->GetValue().GetChar(0); }
    if(mov != NULL) {
        switch(ch_mov->GetSelection()) {
            case 0: *mov = '-'; break;
            case 1: *mov = 'L'; break;
            case 2: *mov = 'R'; break;
            case 3: *mov = 'E'; break;
            default: *mov = '?'; break;
        }
    }
    return;
}

//=============================================================================
Dlg_DelLinkDest::Dlg_DelLinkDest(wxWindow *parent) :
    wxDialog(parent, wxID_ANY, "Delete SNode Link Target", wxDefaultPosition,
     wxDefaultSize, COMMON_FRAME_FLAGS)
{
    wxBoxSizer *vp = new wxBoxSizer(wxVERTICAL),
     *hp = new wxBoxSizer(wxHORIZONTAL);
    wxFlexGridSizer *sp = new wxFlexGridSizer(2);

    vp->Add(sp, 1, wxALIGN_CENTER | wxGROW | wxALL, 3);
    vp->Add(new wxStaticLine(this, wxID_ANY), 0, wxGROW | wxTOP | wxBOTTOM, 5);
    vp->Add(hp, 0, wxALIGN_RIGHT);

    tx_ichar = new wxTextCtrl(this, wxID_ANY, wxEmptyString);
    tx_ichar->SetMaxLength(1);

    sp->Add(new wxStaticText(this, wxID_ANY, "Input char"), 0, wxACV | wxALL, 3);
    sp->Add(tx_ichar, 1, wxGROW | wxACV | wxALL, 3);
    sp->AddGrowableCol(1);

    hp->Add(new wxButton(this, wxID_OK, "&OK"), 0, wxALL, 5);
    hp->Add(new wxButton(this, wxID_CANCEL, "&Cancel"), 0, wxALL, 5);

    SetSizer(vp);
    vp->SetSizeHints(this);
    return;
}

char Dlg_DelLinkDest::result(void) const {
    return tx_ichar->GetValue().GetChar(0);
}

//=============================================================================
Dlg_INodeDest::Dlg_INodeDest(wxWindow *parent) :
    wxDialog(parent, wxID_ANY, "INode reference graph", wxDefaultPosition,
     wxDefaultSize, COMMON_FRAME_FLAGS)
{
    wxBoxSizer *vp = new wxBoxSizer(wxVERTICAL),
     *hp = new wxBoxSizer(wxHORIZONTAL);
    wxFlexGridSizer *sp = new wxFlexGridSizer(2);

    vp->Add(sp, 1, wxALIGN_CENTER | wxGROW | wxALL, 3);
    vp->Add(new wxStaticLine(this, wxID_ANY), 0, wxGROW | wxTOP | wxBOTTOM, 5);
    vp->Add(hp, 0, wxALIGN_RIGHT);

    tx_file = new wxTextCtrl(this, wxID_ANY, wxEmptyString);
    sp->Add(new wxStaticText(this, wxID_ANY, "INode target"), 0, wxACV | wxALL, 3);
    sp->Add(tx_file, 1, wxGROW | wxACV | wxALL, 3);
    sp->AddGrowableCol(1);

    hp->Add(new wxButton(this, wxID_OK, "&OK"), 0, wxALL, 5);
    hp->Add(new wxButton(this, wxID_CANCEL, "&Cancel"), 0, wxALL, 5);

    SetSizer(vp);
    vp->SetSizeHints(this);
    return;
}

wxString Dlg_INodeDest::result(void) const {
    return tx_file->GetValue();
}

//==[ End of file ]============================================================
