/*=============================================================================
Turing Designer
si_sim.cpp - GUI Simulator Engine

  Copyright © Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2005
  -- License restrictions apply (GPL v2)

  This file is part of Turing Designer.
  Turing Desigenr is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; however ONLY version 2 of the License.

  Turing Designer is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program kit; if not, write to:
  Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.

  -- For details see the COPYING file
=============================================================================*/
#include <wx/wxprec.h>
#ifndef WX_PRECOMP
#    include <wx/wx.h>
#endif
#include <wx/statline.h>
#include <stdio.h>
#include <string.h>
#include <libHX.h>
#include "de_genw.hpp"
#include "gr_data.hpp"
#include "si_sim.hpp"
#include "si_tape.hpp"
#include "tu_mach.hpp"
#include "xf_icons.hpp"
#include "xf_main.hpp"
#include "xf_wxsu.hpp"
#define BCOMPRESS(bits)     ((bits + 7) / 8)
#define BIT_CLEAR(map, bit) ((map)[(bit) / 8] &= ~((bit) % 8))
#define BIT_SET(map, bit)   ((map)[(bit) / 8] |= (bit) % 8)
#define BIT_ISSET(map, bit) ((map)[(bit) / 8] & ((bit) % 8))

enum {
    ID_NONE = 0,
    ID_RUN,
    ID_STEP_ONE,
    ID_STEP_UNTIL,
    ID_STEP_MANY,
    ID_STOP,
    ID_RESET,
    ID_TIMER,
};

static const char *reset_msg = "Turing Machine reset to entrypoint.\n";

static size_t count_trailing_zeroes(const char *);

BEGIN_EVENT_TABLE(Simulator, wxPanel)
    EVT_BUTTON(ID_RUN,        Simulator::Run)
    EVT_BUTTON(ID_STEP_ONE,   Simulator::Step_One)
    EVT_BUTTON(ID_STEP_UNTIL, Simulator::Step_Until)
    EVT_BUTTON(ID_STEP_MANY,  Simulator::Step_Many)
    EVT_BUTTON(ID_STOP,       Simulator::Stop)
    EVT_BUTTON(ID_RESET,      Simulator::Reset)
    EVT_TIMER (ID_TIMER,      Simulator::Timer)
END_EVENT_TABLE()

//-----------------------------------------------------------------------------
Simulator::Simulator(wxWindow *parent, TGraph *gr) :
    wxPanel(parent)
{
    graph      = gr;
    tape       = new SimTape(this);
    mach       = new TuMach(graph, tape);
    bp_map     = NULL;
    bp_size    = 0;
    wtimer     = new wxTimer(this, ID_TIMER);
    steps_left = -1;

    // Graphical interface instantiation
    wxBoxSizer *top = new wxBoxSizer(wxVERTICAL);
    top->Add(tape, 0, wxGROW | wxALL, 5);
    top->Add(gen_machview(), 1, wxGROW | wxALL, 5);

    tx_state->AppendText(reset_msg);
    SetSizer(top);
    return;
}

Simulator::~Simulator(void) {
    delete mach;
    delete tape;
    delete bp_map;
    delete wtimer;
    return;
}

void Simulator::Run(wxCommandEvent &ev) {
    unsigned int delay;
    if((delay = get_delay()) <= 0) { return; }
    steps_left = -1;
    wtimer->Start(delay);
    return;
}

void Simulator::Stop(wxCommandEvent &ev) {
    wtimer->Stop();
    return;
}

void Simulator::Step_One(wxCommandEvent &ev) {
    do_tick();
    return;
}

void Simulator::Step_Until(wxCommandEvent &ev) {
    unsigned int delay, steps;
    if((delay = get_delay()) <= 0) { return; }
    if((steps = get_steps()) <= 0) { return; }

    // XXX: need to remove the bp after timer has stopped...
    bp_put(steps);
    steps_left = -1;
    wtimer->Start(delay);
    return;
}

void Simulator::Step_Many(wxCommandEvent &ev) {
    unsigned int delay;
    if((delay = get_delay()) <= 0) { return; }
    steps_left = get_steps();
    wtimer->Start(delay);
    return;
}

void Simulator::Reset(wxCommandEvent &ev) {
    mach->reset();
    tx_state->Clear();
    tx_state->AppendText(reset_msg);
    show_next_step();
    return;
}

void Simulator::show_next_step(void) {
    /* Display the instruction and action the machine would execute next.
    Useful for single step debugging. */
    wxString str, stmp;
    char in, out, mov;
    int cur, next;

    if(!mach->preview(&cur, &in, &out, &mov, &next)) {
        tx_state->AppendText("No state preview available."
         " Do you forgot to load a graph?\n");
        return;
    }
    str.Printf("current=%d", cur);

    if(in != 0) { stmp.Printf(", in='%c'", in); }
    else        { stmp.Printf(", in=--"); }
    str += stmp;

    if(out != 0) { stmp.Printf(", out='%c'", out); }
    else         { stmp.Printf(", out=--"); }
    str += stmp;

    stmp.Printf(", mov=%c, next=%d\n", (mov != 0) ? mov : '-', next);
    str += stmp;
    tx_state->AppendText(str);
    return;
}

//-----------------------------------------------------------------------------
void Simulator::Timer(wxTimerEvent &ev) {
    if(steps_left == -1 || steps_left--) {
        if(do_tick() < 0) {
            // hit breakpoint, error or exit gate
            wtimer->Stop();
        }
    } else {
        wtimer->Stop();
    }
    return;
}

void Simulator::bp_del(unsigned int s) {
    // Delete breakpoint at state S
    if(s >= graph->last_idx() || BCOMPRESS(s + 1) > bp_size) {
        return;
    }
    bp_extend(BCOMPRESS(s + 1));
    BIT_CLEAR(bp_map, s);
    return;
}

bool Simulator::bp_get(unsigned int s) {
    // Check if a breakpoint is set at the given state S.
    if(s >= graph->last_idx() || BCOMPRESS(s + 1) > bp_size) {
        return false;
    }
    return BIT_ISSET(bp_map, s);
}

bool Simulator::bp_put(unsigned int s) {
    // Set a breakpoint at state S (begins counting from zero)
    if(s >= graph->last_idx()) { return false; }
    bp_extend(BCOMPRESS(s + 1));
    BIT_SET(bp_map, s);
    return true;
}

void Simulator::bp_extend(size_t bytes) {
    if(bp_size < bytes) {
        bp_map = reinterpret_cast<unsigned char *>(realloc(bp_map, bytes));
        memset(&bp_map[bp_size], 0, bytes - bp_size);
        bp_size = bytes;
    }
    return;
}

int Simulator::do_tick(void) {
    /* Let the machine do one tick. In case we are already in the EXIT or ERROR
    state, calling tick() will do nothing and return EXIT/ERROR again. */
    int eax;
    if((eax = mach->tick()) >= 0) {
        show_next_step();
        return eax;
    }

    switch(eax) {
        case -1:
            tx_state->AppendText("Reached exit gate. Machine completed.\n");
            tx_state->AppendText(reset_msg);
            // prevent crash when trying to tick() again
            mach->reset();
            break;
        case -2:
            tx_state->AppendText("Reached ERROR state.\n");
            tx_state->AppendText(reset_msg);
            mach->reset();
            break;
        case -3:
            // Should never happen because we do reset() above
            tx_state->AppendText("Machine in an exit/error state."
             " Nothing done.\n");
            break;
        default:
            wxString buf;
            buf.Printf("%s: Machine returned unknown error code\n",
             __PRETTY_FUNCTION__);
            tx_state->AppendText(buf);
            tx_state->AppendText(reset_msg);
            mach->reset();
            break;
    }
    return eax;
}

wxStaticBoxSizer *Simulator::gen_machview(void) {
    wxBoxSizer *hp0, *hp1, *hp2;
    wxStaticBoxSizer *packer;
    wxStaticBox *sbox;

    sbox     = new wxStaticBox(this, wxID_ANY, "Turing Machine");
    packer   = new wxStaticBoxSizer(sbox, wxVERTICAL);
    tx_state = new wxTextCtrl(this, wxID_ANY, "", wxDefaultPosition,
               wxDefaultSize, wxTE_MULTILINE);

    hp0      = new wxBoxSizer(wxHORIZONTAL);
    hp1      = new wxBoxSizer(wxHORIZONTAL);
    hp2      = new wxBoxSizer(wxHORIZONTAL);
    tx_step  = new wxTextCtrl(this, wxID_ANY, "", wxDefaultPosition,
               wxDefaultSize, wxTE_RIGHT);
    //cb_step  = new wxCheckBox(this, wxID_ANY, "Show steps");
    //cb_step->SetValue(true);
    tx_delay = new wxTextCtrl(this, wxID_ANY, "250", wxDefaultPosition,
               wxDefaultSize, wxTE_RIGHT);

    hp1->Add(new wxStaticText(this, wxID_ANY, "Run:"), 0, wxACV | wxALL, 3);
    hp1->Add(new GWButton(this, ID_RUN, *__icon_bt_start, "Run"), 0, wxACV | wxALL, 3);
    hp1->Add(new GWButton(this, ID_STEP_ONE, *__icon_bt_pause, "Step One"), 0, wxACV | wxALL, 3);
    hp1->Add(new GWButton(this, ID_STOP, *__icon_bt_stop, "Stop"), 0, wxACV | wxALL, 3);
    hp1->Add(new GWButton(this, ID_RESET, *__icon_bt_reset, "Reset"), 0, wxACV | wxALL, 3);
    //XXX//hp1->Add(new wxButton(this, ID_STEP_UNTIL, "until >"), 0, wxACV | wxALL, 3);
    hp1->Add(tx_step, 0, wxACV | wxALL, 3);
    hp1->Add(new wxButton(this, ID_STEP_MANY, "< steps"), 0, wxACV | wxALL, 3);
    hp1->Add(new wxStaticLine(this, wxID_ANY, wxDefaultPosition, wxDefaultSize, wxLI_VERTICAL), 0, wxGROW | wxACV | wxLEFT | wxRIGHT, 5);

    //hp2->Add(cb_step, 0, wxACV | wxALL, 3);
    hp2->Add(tx_delay, 0, wxACV | wxALL, 3);
    hp2->Add(new wxStaticText(this, wxID_ANY, "ms"), 0, wxACV | wxALL, 3);

    hp0->Add(hp1, 1, wxALIGN_BOTTOM);
    hp0->Add(hp2, 0, wxALIGN_BOTTOM);

    packer->Add(tx_state, 1, wxALL | wxALIGN_LEFT | wxGROW, 3);
    packer->Add(hp0, 0, wxGROW | wxALL | wxALIGN_RIGHT, 0);
    return packer;
}

int Simulator::get_delay(void) {
    char *err = NULL;
    unsigned int delay = strtoul(tx_delay->GetValue(), &err, 0);

    if(*err != '\0') {
        GWMessage(this, "Step Until", "Invalid delay value", wxOK).ShowModal();
        return -1;
    }

    return delay;
}

int Simulator::get_steps(void) {
    char *err = NULL;
    unsigned int steps = strtoul(tx_step->GetValue(), &err, 0);

    if(*err != '\0') {
        GWMessage(this, "Step Until", "Invalid value in that textfield"
         " next to \"Step Until\"!", wxOK).ShowModal();
        return -1;
    }

    return steps;
}

//==[ End of file ]============================================================
