/*=============================================================================
Turing Designer
si_tape.cpp - GUI Tape

  Copyright © Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2005
  -- License restrictions apply (GPL v2)

  This file is part of Turing Designer.
  Turing Desigenr is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; however ONLY version 2 of the License.

  Turing Designer is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program kit; if not, write to:
  Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.

  -- For details see the COPYING file
=============================================================================*/
#include <errno.h>
#include <string.h>
#include <libHX.h>
#include <wx/wxprec.h>
#ifndef WX_PRECOMP
#    include <wx/wx.h>
#endif
#include "si_sim.hpp"
#include "si_tape.hpp"

enum {
    ID_NONE = 0,
    ID_INIT,
    ID_MOVL,
    ID_MOVR,
};

static const char *head_uninit = "(Head uninitialized, prepend '&&' to mark)";

//-----------------------------------------------------------------------------
BEGIN_EVENT_TABLE(SimTape, wxPanel)
    EVT_BUTTON(ID_INIT, SimTape::Initialize)
    EVT_BUTTON(ID_MOVL, SimTape::Movl)
    EVT_BUTTON(ID_MOVR, SimTape::Movr)
END_EVENT_TABLE()

SimTape::SimTape(Simulator *super) :
    wxPanel(super)
    // don't call TuTape ctor
{
    wxStaticBoxSizer *hp0;
    wxBoxSizer *vp1, *vp2, *vp3;
    wxStaticBox *sbox;
    wxButton *button;
    wxChoice *machtype;
    wxFont tt(14, wxTELETYPE, wxNORMAL, wxBOLD, false, "courier");

    sbox = new wxStaticBox(this, wxID_ANY, "Tape");
    hp0  = new wxStaticBoxSizer(sbox, wxHORIZONTAL);

    vp1 = new wxBoxSizer(wxVERTICAL);
    tx_init = new wxTextCtrl(this, wxID_ANY, "3&33222000111333");
    tx_data = new wxStaticText(this, wxID_ANY, "(Tape uninitialized)");
    tx_head = new wxStaticText(this, wxID_ANY, head_uninit);
    tx_data->SetFont(tt);
    tx_head->SetFont(tt);
    vp1->Add(tx_init, 0, wxGROW | wxALL, 3);
    vp1->Add(tx_data, 0, wxGROW | wxALL, 3);
    vp1->Add(tx_head, 0, wxGROW | wxALL, 3);

    vp2 = new wxBoxSizer(wxVERTICAL);
    vp2->Add(new wxButton(this, ID_INIT, "Initialize"), 0, wxALL, 3);

    vp3 = new wxBoxSizer(wxHORIZONTAL);
    vp3->Add(new wxButton(this, ID_MOVL, "<<", wxDefaultPosition, wxDefaultSize, wxBU_EXACTFIT), 1, wxALL, 3);
    vp3->Add(new wxButton(this, ID_MOVR, ">>", wxDefaultPosition, wxDefaultSize, wxBU_EXACTFIT), 1, wxALL, 3);
    vp2->Add(vp3, 0, wxALIGN_CENTER);

    hp0->Add(vp1, 1, wxGROW);
    hp0->Add(vp2, 0, wxGROW);
    SetSizer(hp0);

    str_data = new wxString();
    str_head = new wxString();
    head_pos = -1;
    this->super = super;
    return;
}

SimTape::~SimTape(void) {
    delete str_data;
    delete str_head;
    return;
}

char &SimTape::operator *(void) const {
    // Get the character under the pointer
    return str_data->GetWritableChar(head_pos);
}

char &SimTape::operator [](size_t i) {
    /* Get the character at position I, expanding the GUI text strings if
    necessary ("auto-vivification"). */
    extend(i + 1);
    return str_data->GetWritableChar(i);
}

int SimTape::operator --(void) {
    /* Moves the pointer to the left. Note that a SimTape is only infinite to
    the right.
    Returns: the new pointer position (counting from zero), or
             negative to indicate failure
    */
    if(head_pos <= 0) { return -1; }
    str_head->GetWritableChar(head_pos)   = ' ';
    str_head->GetWritableChar(--head_pos) = '^';
    update();
    return head_pos;
}

int SimTape::operator ++(void) {
    // Moves the pointer to the right
    if(head_pos < 0) { return -1; }
    extend(head_pos + 2); // future position plus 1: makes 2
    str_head->GetWritableChar(head_pos)   = ' ';
    str_head->GetWritableChar(++head_pos) = '^';
    update();
    return head_pos;
}

void SimTape::update(void) {
    tx_data->SetLabel(*str_data);
    tx_head->SetLabel(*str_head);
    return;
}

//-----------------------------------------------------------------------------
void SimTape::Initialize(wxCommandEvent &ev) {
    wxString &rd = *str_data;
    rd = tx_init->GetValue();

    if((head_pos = rd.Find('&')) >= 0) {
        str_head->Truncate(0);
        rd = rd(0, head_pos) + rd(head_pos + 1, rd.Length() - head_pos - 1);
        extend(head_pos + 1);
        str_head->GetWritableChar(head_pos) = '^';
    }

    tx_data->SetLabel(rd);
    update();
    super->show_next_step();
    return;
}

void SimTape::Movl(wxCommandEvent &ev) {
    --*this;
    return;
}

void SimTape::Movr(wxCommandEvent &ev) {
    ++*this;
    return;
}

void SimTape::extend(size_t n) {
    /* Extend the tape to the right side, if necessary. (Extend the TAPE and
    HEAD strings.) */
    int l;
    if((l = str_data->Length()) < n) { str_data->Pad(n - l); }
    if((l = str_head->Length()) < n) { str_head->Pad(n - l); }
    return;
}

//==[ End of file ]============================================================
