/*=============================================================================
Turing Designer
tu_mach.cpp - UI-neutral Turing Machine

  Copyright © Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2005
  -- License restrictions apply (GPL v2)

  This file is part of Turing Designer.
  Turing Desigenr is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; however ONLY version 2 of the License.

  Turing Designer is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program kit; if not, write to:
  Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.

  -- For details see the COPYING file
=============================================================================*/
#include <ctype.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include "gr_data.hpp"
#include "tu_mach.hpp"
#include "tu_tape.hpp"

TuMach::TuMach(TGraph *gr, TuTape *tp) {
    graph = gr;
    tape  = tp;
    reset();
    return;
}

int TuMach::run(void) {
    int eax;
    while((eax = tick()) >= 0);
    return eax;
}

bool TuMach::preview(int *curp, char *inp, char *outp, char *movp, int *next) {
    SNode *node = dynamic_cast<SNode *>((*graph)[current]);
    if(node == NULL) {
        /* I,L-Nodes are "atomic" from outside TuMach... remember? We should
        not get here, _because_ they are transparently ran across. */
        return false;
    }
    ATarget *target;
    *curp  = current;
    *inp   = **tape;
    target = node->get(*inp);
    if(target == NULL) { target = node->get(0); }
    if(target == NULL) {
        *outp = 0;
        *movp = 0;
        *next = -1;
    } else {
        *outp = target->out;
        *movp = toupper(target->mov);
        *next = target->next;
    }
    return true;
}

int TuMach::tick(void) {
    /* Advance by exactly one state. (Not entirely true, but see below.)
    Returns: The state number we are at after the tick.
    */
    if(current < 0) { return -3; }
    VNode *ptr = (*graph)[current];

    switch(ptr->type()) {
        case TYPE_SNODE: { // that's the red bubbles
            SNode *node = dynamic_cast<SNode *>(ptr);
            char i = **tape;
            ATarget *target = node->get(i);

            if(target == NULL) {
                // Nice feature: A "default" target
                target = node->get(0);
            }
            if(target == NULL) {
                // No targets found, must be the exit node!
                return -1;
            }
            if(target->out != '\0') {
                // Nice feature 2: Keep value
                **tape = target->out;
            }
            switch(toupper(target->mov)) {
                // Hardcoded in Dlg_LinkDest::
                case 'L': --*tape; break;
                case 'R': ++*tape; break;
                case 'E': return -2; break;
            }
            tape->update();
            current = target->next;
            break;
        }
        case TYPE_INODE: {
            // This is all the magic of stacked automatons
            INode *node = dynamic_cast<INode *>(ptr);
            TuMach tm(node->graph, tape); // automatically destroyed
            tm.run();
            current = node->next;
            break;
        }
        case TYPE_LNODE: {
            // Simple relay nodes...
            current = dynamic_cast<LNode *>(ptr)->next;
            break;
        }
    }

    if(current < 0) { return current; }

    if((ptr = (*graph)[current]) == NULL) {
        fprintf(stderr, "%s: Someone broke the graph structure!\n",
         __PRETTY_FUNCTION__);
    }
    if(ptr->type() == TYPE_SNODE) {
        return current;
    }
    /* The Drawpad can't yet "enter" into stacked graphs, so run INodes
    and LNodes atomically. Let's hope GCC has tail-recursing if you have A LOT
    of I/L-Nodes. :-) */
    return tick();
}

void TuMach::reset(void) {
    current = (graph->entrypoint >= 0) ? graph->entrypoint : 0;
    return;
}

//==[ End of file ]============================================================
