/*=============================================================================
Turing Designer
xf_main.cpp - GUI MainWindow

  Copyright © Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2005
  -- License restrictions apply (GPL v2)

  This file is part of Turing Designer.
  Turing Desigenr is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; however ONLY version 2 of the License.

  Turing Designer is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program kit; if not, write to:
  Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.

  -- For details see the COPYING file
=============================================================================*/
#include <errno.h>
#include <stdio.h>
#include <wx/wxprec.h>
#ifndef WX_PRECOMP
#    include <wx/wx.h>
#endif
#include <wx/mstream.h>
#include <wx/statline.h>
#include "de_data.hpp"
#include "de_pad.hpp"
#include "de_genw.hpp"
#include "gr_data.hpp"
#include "si_sim.hpp"
#include "xf_icons.hpp"
#include "xf_main.hpp"
#include "xf_wxsu.hpp"
#include "xf_xml.hpp"

enum {
    ID_NONE,

    ID_ENTRY,

    ID_MARK,
    ID_SNODE,
    ID_INODE,
    ID_LNODE,
    ID_LINK,
    ID_LINKD,

    ID_RUN,
    ID_STEP,
    ID_STOP,
    ID_RESET,

    ID_ROLLDN,
    ID_ROLLUP,
    ID_ROLLEQ,
};

static const wxColour CL_DarkGreen(0, 128, 0);
static const char *File_filter = "Turing Designer Transition Graph files (*.xml)|*.xml";

//-----------------------------------------------------------------------------
BEGIN_EVENT_TABLE(MainWindow, wxFrame)
    // File menu
    EVT_MENU(wxID_NEW,    MainWindow::File_New)
    EVT_MENU(wxID_OPEN,   MainWindow::File_Open)
    EVT_MENU(wxID_SAVE,   MainWindow::File_Save)
    EVT_MENU(wxID_SAVEAS, MainWindow::File_SaveAs)
    EVT_MENU(wxID_EXIT,   MainWindow::File_Exit)

    // Edit menu
    EVT_MENU(wxID_DELETE, MainWindow::Edit_Delete)
    EVT_MENU(ID_ENTRY,    MainWindow::Edit_Entrypoint)

    // Tool menu
    EVT_MENU(ID_MARK,  MainWindow::Tool_Mark)
    EVT_MENU(ID_SNODE, MainWindow::Tool_SNode)
    EVT_MENU(ID_INODE, MainWindow::Tool_INode)
    EVT_MENU(ID_LNODE, MainWindow::Tool_LNode)
    EVT_MENU(ID_LINK,  MainWindow::Tool_Link)
    EVT_MENU(ID_LINKD, MainWindow::Tool_DelLink)

    // Tool buttons (same callbacks as tool menu)
    EVT_BUTTON(ID_MARK,  MainWindow::Tool_Mark)
    EVT_BUTTON(ID_SNODE, MainWindow::Tool_SNode)
    EVT_BUTTON(ID_INODE, MainWindow::Tool_INode)
    EVT_BUTTON(ID_LNODE, MainWindow::Tool_LNode)
    EVT_BUTTON(ID_LINK,  MainWindow::Tool_Link)
    EVT_BUTTON(ID_LINKD, MainWindow::Tool_DelLink)

    // Run menu
    EVT_MENU(ID_RUN,   MainWindow::Run_Run)
    EVT_MENU(ID_STEP,  MainWindow::Run_Step)
    EVT_MENU(ID_STOP,  MainWindow::Run_Stop)
    EVT_MENU(ID_RESET, MainWindow::Run_Reset)

    // About menu
    EVT_MENU(wxID_ABOUT, MainWindow::About)

    // Debug buttons (same callbacks as debug menu)
    EVT_BUTTON(ID_RUN,   MainWindow::Run_Run)
    EVT_BUTTON(ID_STOP,  MainWindow::Run_Stop)
    EVT_BUTTON(ID_STEP,  MainWindow::Run_Step)
    EVT_BUTTON(ID_RESET, MainWindow::Run_Reset)

    // Separator
    EVT_BUTTON(ID_ROLLDN, MainWindow::Roll_Down)
    EVT_BUTTON(ID_ROLLUP, MainWindow::Roll_Up)
    EVT_BUTTON(ID_ROLLEQ, MainWindow::Roll_Equal)
END_EVENT_TABLE()

MainWindow::MainWindow(const char *title, const wxPoint &pos,
 const wxSize &size, TGraph *tg) :
    wxFrame(NULL, wxID_ANY, title, pos, size)
{
    // Data
    graph    = (tg != NULL) ? tg : new TGraph();
    filename = NULL;

    // Menu bar
    GWMenu *menu_file = new GWMenu();
    menu_file->Append_WIcon(wxID_NEW,    "&New"     "\t" "Ctrl-N", *__icon_bt_new);
    menu_file->Append_WIcon(wxID_OPEN,   "&Open"    "\t" "Ctrl-O", *__icon_bt_open);
    menu_file->Append_WIcon(wxID_SAVE,   "&Save"    "\t" "Ctrl-S", *__icon_bt_save);
    menu_file->Append_WIcon(wxID_SAVEAS, "Save &as" "\t" "F12",    *__icon_bt_savea);
    menu_file->AppendSeparator();
    menu_file->Append(wxID_EXIT,   "E&xit"    "\t" "Alt-F4");

    GWMenu *menu_edit = new GWMenu();
    menu_edit->Append_WIcon(wxID_DELETE, "&Delete selected node(s)", *__icon_bt_erase);
    menu_edit->Append(ID_ENTRY, "Set &Entrypoint");

    GWMenu *menu_tool = new GWMenu();
    menu_tool->Append_WIcon(ID_MARK,  "&Select tool",  *__icon_bt_sel);
    menu_tool->Append_WIcon(ID_SNODE, "Insert &SNode", *__icon_bt_snode);
    menu_tool->Append_WIcon(ID_INODE, "Insert &INode", *__icon_bt_inode);
    menu_tool->Append_WIcon(ID_LNODE, "Insert &LNode", *__icon_bt_lnode);
    menu_tool->Append_WIcon(ID_LINK,  "Insert Lin&k",  *__icon_bt_arrow);
    menu_tool->AppendSeparator();
    menu_tool->Append_WIcon(ID_LINKD, "Delete link",   *__icon_bt_arrwd);

    GWMenu *menu_run = new GWMenu();
    menu_run->Append_WIcon(ID_RUN,   "Run"      "\t" "F9", *__icon_bt_start);
    menu_run->Append_WIcon(ID_STEP,  "Step One" "\t" "F8", *__icon_bt_pause);
    menu_run->Append_WIcon(ID_STOP,  "Stop"     "\t" "F5", *__icon_bt_stop);
    menu_run->Append_WIcon(ID_RESET, "Reset",              *__icon_bt_reset);

    wxMenu *menu_help = new wxMenu();
    menu_help->Append(wxID_ABOUT, "&About" "\t" "F1");

    wxMenuBar *menu_bar = new wxMenuBar(wxMB_DOCKABLE);
    menu_bar->Append(menu_file, "&File");
    menu_bar->Append(menu_edit, "&Edit");
    menu_bar->Append(menu_tool, "&Tools");
    menu_bar->Append(menu_run,  "&Run");
    menu_bar->Append(menu_help, "&Help");

    SetMenuBar(menu_bar);

    // Main area
    packer = new wxBoxSizer(wxVERTICAL);
    drw = new wxBoxSizer(wxHORIZONTAL);
    pad = new Drawpad(this, graph);
    drw->Add(create_drawbar(), 0, wxGROW);
    drw->Add(pad, 1, wxGROW | wxALL, 3);

    wxBoxSizer *hp0 = new wxBoxSizer(wxHORIZONTAL);
    hp0->Add(new wxStaticLine(this, wxID_ANY), 1, wxGROW | wxACV | wxALL, 5);
    hp0->Add(new GWButton(this, ID_ROLLDN, *__icon_bt_roldn, "Expand Designer"), 0, wxACV | wxALL, 2);
    hp0->Add(new GWButton(this, ID_ROLLUP, *__icon_bt_rolup, "Expand Simulator"), 0, wxACV | wxALL, 2);
    hp0->Add(new GWButton(this, ID_ROLLEQ, *__icon_bt_roleq, "Equal sizes"), 0, wxACV | wxALL, 2);
    hp0->Add(new wxStaticLine(this, wxID_ANY), 1, wxGROW | wxACV | wxALL, 5);

    sim = new Simulator(this, graph);
    packer->Add(drw, 1, wxGROW);
    packer->Add(hp0, 0, wxGROW);
    packer->Add(sim, 1, wxGROW);

    // Status bar
    int widths[] = {0, -1, 80};
    CreateStatusBar(3);
    SetStatusWidths(3, widths);
    SetStatusText("by Jan Engelhardt"
     " <jengelh [at] linux01 gwdg de>, 2005", 1);

    // Final window size
    this->SetSizer(packer);
    packer->SetSizeHints(this);

    wxSize cursize = GetSize();
    cursize.IncTo(wxSize(800, 600));
    SetSize(cursize);
    return;
}

MainWindow::~MainWindow(void) {
    delete graph;
    return;
}

//-----------------------------------------------------------------------------
void MainWindow::File_New(wxCommandEvent &ev) {
    new_mw();
    return;
}

void MainWindow::File_Open(wxCommandEvent &ev) {
    wxFileDialog dialog(this, "File > Open", wxEmptyString, wxEmptyString,
      File_filter, wxOPEN | wxFILE_MUST_EXIST | wxCHANGE_DIR);
    if(dialog.ShowModal() != wxID_OK) {
        return;
    }

    const char *fn = dialog.GetFilename();
    TGraph *new_graph;

    if((new_graph = xg_load(fn)) == NULL) {
        wxString buf;
        buf.Printf("Error opening %s: %s", fn, strerror(errno));
        GWMessage(this, "File > Open", buf, wxOK).ShowModal();
        return;
    }

    new_graph->modified = false;
    MainWindow *mw = new_mw(new_graph);
    if(!graph->modified) {
        mw->Move(GetPosition());
        Close();
    }
    return;
}

void MainWindow::File_Save(wxCommandEvent &ev) {
    save_graph(false);
    return;
}

void MainWindow::File_SaveAs(wxCommandEvent &ev) {
    save_graph_as(true);
    return;
}

void MainWindow::File_Exit(wxCommandEvent &ev) {
    Close(true);
    return;
}

void MainWindow::Edit_Delete(wxCommandEvent &ev) {
    pad->Delete(ev);
    return;
}

void MainWindow::Edit_Entrypoint(wxCommandEvent &ev) {
    pad->ext_action(DWEXT_ENTRYPT);
    return;
}

void MainWindow::Tool_Mark(wxCommandEvent &ev) {
    pad->ext_action(DWEXT_MARK);
    return;
}

void MainWindow::Tool_SNode(wxCommandEvent &ev) {
    pad->ext_action(DWEXT_SNODE);
    return;
}

void MainWindow::Tool_INode(wxCommandEvent &ev) {
    pad->ext_action(DWEXT_INODE);
    return;
}

void MainWindow::Tool_LNode(wxCommandEvent &ev) {
    pad->ext_action(DWEXT_LNODE);
    return;
}

void MainWindow::Tool_Link(wxCommandEvent &ev) {
    pad->ext_action(DWEXT_LINK);
    return;
}

void MainWindow::Tool_DelLink(wxCommandEvent &ev) {
    pad->ext_action(DWEXT_LINKD);
    return;
}

void MainWindow::Run_Run(wxCommandEvent &ev) {
    sim->Run(ev);
    return;
}

void MainWindow::Run_Step(wxCommandEvent &ev) {
    sim->Step_One(ev);
    return;
}

void MainWindow::Run_Stop(wxCommandEvent &ev) {
    sim->Stop(ev);
    return;
}

void MainWindow::Run_Reset(wxCommandEvent &ev) {
    sim->Reset(ev);
    return;
}

void MainWindow::About(wxCommandEvent &ev) {
    GWMessage(this, "About",
      "Turing Designer v0.50\n"
      "by Jan Engelhardt\n"
      "<jengelh [at] linux01 gwdg de>\n"
      "released under the GNU Lesser\n"
      "General Public License v2.0\n"
      "© 2005\n",
      wxOK, wxALIGN_CENTER).ShowModal();
    return;
}

void MainWindow::Roll_Up(wxCommandEvent &ev) {
    packer->Show(drw, false);
    packer->Show(sim, true);
    packer->Layout();
    return;
}

void MainWindow::Roll_Down(wxCommandEvent &ev) {
    packer->Show(drw, true);
    packer->Show(sim, false);
    packer->Layout();
    return;
}

void MainWindow::Roll_Equal(wxCommandEvent &ev) {
    packer->Show(drw, true);
    packer->Show(sim, true);
    packer->Layout();
    return;
}

int MainWindow::check_modified(const char *title) {
    if(!graph->modified) { return 1; }

    GWMessage dialog(this, title, "The current graph has been modified. Save changes?", wxYES_NO | wxCANCEL);
    int rv = dialog.ShowModal();
    if(rv == wxID_CANCEL || (rv == wxID_YES && save_graph(0) != wxID_OK)) {
        return 0;
    }

    return 1;
}

wxBoxSizer *MainWindow::create_drawbar(void) {
    wxBoxSizer *v = new wxBoxSizer(wxVERTICAL);
    v->Add(new GWButton(this, ID_MARK,  *__icon_bt_sel  , "Select"), 0, wxALL, 3);
    v->Add(new GWButton(this, ID_SNODE, *__icon_bt_snode, "Insert new SNode"), 0, wxALL, 3);
    v->Add(new GWButton(this, ID_INODE, *__icon_bt_inode, "Insert new INode"), 0, wxALL, 3);
    v->Add(new GWButton(this, ID_LNODE, *__icon_bt_lnode, "Insert new LNode"), 0, wxALL, 3);
    v->Add(new GWButton(this, ID_LINK,  *__icon_bt_arrow, "Insert new Link"), 0, wxALL, 3);
    v->Add(new GWButton(this, ID_LINKD, *__icon_bt_arrwd, "Delete link"), 0, wxALL, 3);
    v->Add(new wxStaticLine(this, wxID_ANY), 0, wxGROW | wxALL, 3);
    return v;
}

int MainWindow::save_graph(bool force) {
    int eax;
    if(filename == NULL) {
        return save_graph_as(force);
    }

    if(!graph->modified && !force) {
        // "Fast save". To force a rewrite, use "Save as" with same filename.
        return wxID_OK;
    }

    if((eax = xg_save(filename, graph)) <= 0) {
        // Take care not to overwrite errno until the snprintf()
        wxString b;
        b.Printf("Error writing to %s: %s\n"
         "Press OK to continue (and possibly discard\n"
         "changes if you selected e.g. \"File > New\")",
         filename, strerror(errno));

        GWMessage dialog(this, "File > Save", b, wxOK | wxCANCEL);
        if(dialog.ShowModal() == wxID_CANCEL) { return wxID_CANCEL; }
    } else {
        graph->modified = false;
    }

    return wxID_OK;
}

int MainWindow::save_graph_as(bool force) {
    wxFileDialog dialog(this, "File > Save as", wxEmptyString,
     (filename == NULL) ? "" : filename, File_filter,
     wxSAVE | wxOVERWRITE_PROMPT | wxCHANGE_DIR);

    if(dialog.ShowModal() == wxID_CANCEL) {
        return wxID_CANCEL;
    }

    HX_strclone(&filename, dialog.GetFilename());
    return save_graph(force);
}

//==[ End of file ]============================================================
