/*=============================================================================
Turing Designer
xf_xml.cpp - Load/Save routines (GUI-specific)

  Copyright © Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2005
  -- License restrictions apply (GPL v2)

  This file is part of Turing Designer.
  Turing Desigenr is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; however ONLY version 2 of the License.

  Turing Designer is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program kit; if not, write to:
  Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.

  -- For details see the COPYING file
=============================================================================*/
#include <libxml/parser.h>
#include <sys/types.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <libHX.h>
#include "de_data.hpp"
#include "gr_data.hpp"
#include "xf_xml.hpp"
#define NULL_ZSTR(ptr) (((ptr) == NULL) ? "" : (ptr))

/* Funky C++ requires cast between unsigned ptr and signed ptr... sigh.
Makes it a pain to code to libxml. */
typedef xmlChar *_XC;
typedef const char *_CC;
typedef char *_CP;

static void xg_load_S(TGraph &, xmlNode *);
static void xg_load_I(TGraph &, xmlNode *);
static void xg_load_L(TGraph &, xmlNode *);
template<class Object> static Object *xg_load(TGraph &, xmlNode *, Object *);
static void xg_save(FILE *, size_t, XSNode *, TGraph *);
static void xg_save(FILE *, size_t, XINode *, TGraph *);
static void xg_save(FILE *, size_t, XLNode *, TGraph *);

//-----------------------------------------------------------------------------
TGraph *xg_load(const char *file) {
    xmlNode *ptr;
    xmlDoc *doc;
    TGraph *ret;

    xmlKeepBlanksDefault(0);
    if((doc = xmlParseFile(file)) == NULL) { return 0; }
    if((ptr = xmlDocGetRootElement(doc)) == NULL ||
     strcmp((_CC)ptr->name, "tgraph") != 0) {
        xmlFreeDoc(doc);
        xmlCleanupParser();
        return 0;
    }

    ptr = ptr->children;
    ret = new TGraph();
    while(ptr != NULL) {
        if(strcmp((_CC)ptr->name, "snode") == 0) { xg_load_S(*ret, ptr); }
        else if(strcmp((_CC)ptr->name, "lnode") == 0) { xg_load_L(*ret, ptr); }
        else if(strcmp((_CC)ptr->name, "inode") == 0) { xg_load_I(*ret, ptr); }
        ptr = ptr->next;
    }

    return ret;
}

int xg_save(const char *file, TGraph *graph) {
    unsigned int max = graph->last_idx() + 1;
    FILE *fp;
    if((fp = fopen(file, "w")) == NULL) { return -errno; }
    fprintf(fp, "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n" "<tgraph>\n");

    for(unsigned int i = 0; i < max; ++i) {
        VNode *node = (*graph)[i];
        if(node == NULL) { continue; }

        switch(node->type()) {
            case TYPE_SNODE:
                xg_save(fp, i, dynamic_cast<XSNode *>(node), graph);
                break;
            case TYPE_INODE:
                xg_save(fp, i, dynamic_cast<XINode *>(node), graph);
                break;
            case TYPE_LNODE:
                xg_save(fp, i, dynamic_cast<XLNode *>(node), graph);
                break;
            default:
                fprintf(stderr, "%s: Illegal node type\n", __PRETTY_FUNCTION__);
                break;
        }
    }
    fprintf(fp, "</tgraph>\n");
    fclose(fp);
    return 1;
}

//-----------------------------------------------------------------------------
static void xg_load_S(TGraph &ret, xmlNode *ptr) {
    XSNode *nd  = xg_load(ret, ptr, reinterpret_cast<XSNode *>(NULL));
    xmlNode *cd = ptr->children;

    while(cd != NULL) {
        if(strcmp((_CC)cd->name, "target") != 0) {
            cd = cd->next;
            continue;
        }

        char *s_in  = (_CP)xmlGetProp(cd, (_XC)"in"),
            *s_out  = (_CP)xmlGetProp(cd, (_XC)"out"),
            *s_mov  = (_CP)xmlGetProp(cd, (_XC)"mov"),
            *s_next = (_CP)xmlGetProp(cd, (_XC)"next");
        char i_in   = (s_in   != NULL) ? strtol(s_in,   NULL, 0) : 0,
             i_out  = (s_out  != NULL) ? strtol(s_out,  NULL, 0) : 0,
             i_mov  = (s_mov  != NULL) ? *s_mov : 0,
             i_next = (s_next != NULL) ? strtol(s_next, NULL, 0) : -1;

        if(i_next >= 0) { ret(i_next); } // extend
        nd->put(i_in, i_out, i_mov, i_next);
        free(s_in);
        free(s_out);
        free(s_mov);
        free(s_next);
        cd = cd->next;
    }

    return;
}

static void xg_load_I(TGraph &ret, xmlNode *ptr) {
    XINode *nd   = xg_load(ret, ptr, reinterpret_cast<XINode *>(NULL));
    char *s_next = (_CP)xmlGetProp(ptr, (_XC)"next");
    nd->ext      = (_CP)xmlGetProp(ptr, (_XC)"graph");
    nd->next     = (s_next != NULL) ? strtol(s_next, NULL, 0) : -1;
    nd->graph    = xg_load(nd->ext);
    return;
}

static void xg_load_L(TGraph &ret, xmlNode *ptr) {
    XLNode *nd   = xg_load(ret, ptr, reinterpret_cast<XLNode *>(NULL));
    char *s_next = (_CP)xmlGetProp(ptr, (_XC)"next");
    nd->next     = (s_next != NULL) ? strtol(s_next, NULL, 0) : -1;
    return;
}

template<class Object> static Object *xg_load(TGraph &ret, xmlNode *ptr,
 Object *dummy)
{
    char *s_id = (_CP)xmlGetProp(ptr, (_XC)"id"),
         *s_xp = (_CP)xmlGetProp(ptr, (_XC)"x"),
         *s_yp = (_CP)xmlGetProp(ptr, (_XC)"y");
    int i_xp   = (s_xp != NULL) ? strtol(s_xp, NULL, 0) : 0,
        i_yp   = (s_yp != NULL) ? strtol(s_yp, NULL, 0) : 0;
    Object *nd = new Object(i_xp, i_yp);

    if(s_id != NULL) {
        unsigned int id = strtoul(s_id, NULL, 0);
        ret(id); // extend space
        ret[id] = nd;
    } else {
        ret.add(nd);
    }

    free(s_id);
    free(s_xp);
    free(s_yp);
    return nd;
}

static void xg_save(FILE *fp, unsigned int i, XSNode *n, TGraph *graph) {
    struct HXbtree_node *w;
    void *travp = HXbtrav_init(n->jmp, NULL);

    fprintf(fp, "  <snode id=\"%u\" x=\"%d\" y=\"%d\">\n", i, n->x, n->y);
    while((w = HXbtraverse(travp)) != NULL) {
        ATarget *target = reinterpret_cast<ATarget *>(w->data);
        fprintf(fp, "    <target in=\"%d\" out=\"%d\" mov=\"%c\""
          " next=\"%d\" />\n",
          reinterpret_cast<int>(w->key), target->out, target->mov,
          target->next
        );
    }
    fprintf(fp, "  </snode>\n");
    return;
}

static void xg_save(FILE *fp, unsigned int i, XINode *n, TGraph *graph) {
    fprintf(fp,
      "  <inode id=\"%u\" x=\"%d\" y=\"%d\" graph=\"%s\" next=\"%d\" />\n",
      i, n->x, n->y, NULL_ZSTR(n->ext), n->next
    );
    return;
}

static void xg_save(FILE *fp, unsigned int i, XLNode *n, TGraph *graph) {
    fprintf(fp, "  <lnode id=\"%u\" x=\"%d\" y=\"%d\" next=\"%d\" />\n",
     i, n->x, n->y, n->next);
    return;
}

//==[ End of file ]============================================================
